@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1996-2012 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2016 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -86,7 +86,7 @@ public:
 
    void SetDate(time_t d,int p=0);
    void SetSize(off_t s);
-   void SetEntitySize(off_t s) { e_size=s; }
+   void SetEntitySize(off_t s) { if(!ascii) e_size=s; }
 
    void DontCopyDate() { do_set_date=false; }
    void DontVerify() { do_verify=false; }
@@ -174,6 +174,9 @@ private:
    int put_buf;
    off_t put_eof_pos;
 
+   off_t high_watermark;
+   Timer high_watermark_timeout;
+
    Time start_time;
    Time end_time;
 
@@ -254,7 +257,9 @@ public:
    bool WritePending() { return put && put->WritePending(); }
 
    void LogTransfer();
-   static SMTaskRef<Log> transfer_log;
+   static Ref<Log> transfer_log;
+
+   static const char *TempFileName(const char *file);
 };
 
 class FileVerificator : public SMTask
