@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1996-2012 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2016 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -50,10 +50,16 @@ void  FileInfo::Merge(const FileInfo& f)
 {
    if(strcmp(basename_ptr(name),basename_ptr(f.name)))
       return;
-// int sim=defined&f.defined;
-   int dif=(~defined)&f.defined;
-   if(dif&MODE)
+   MergeInfo(f,~defined);
+}
+void  FileInfo::MergeInfo(const FileInfo& f,unsigned dif)
+{
+   dif&=f.defined;
+   if(dif&MODE) {
       SetMode(f.mode);
+      if(mode!=SYMLINK && mode!=REDIRECT)
+	 symlink.unset();
+   }
    if(dif&DATE || (defined&DATE && f.defined&DATE && f.date.ts_prec<date.ts_prec))
       SetDate(f.date,f.date.ts_prec);
    if(dif&SIZE)
@@ -115,9 +121,17 @@ void FileSet::Sub(int i)
    if(ind>i)
       ind--;
 }
+FileInfo *FileSet::Borrow(int i)
+{
+   FileInfo *fi=files[i].borrow();
+   Sub(i);
+   return fi;
+}
 
 void FileSet::Merge(const FileSet *set)
 {
+   if(!set)
+      return;
    for(int i=0; i<set->fnum; i++)
    {
       const Ref<FileInfo>& fi=set->files[i];
@@ -284,12 +298,18 @@ void FileSet::SubtractAny(const FileSet *set)
 
 void FileSet::SubtractNotIn(const FileSet *set)
 {
+   if(!set) {
+      Empty();
+      return;
+   }
    for(int i=0; i<fnum; i++)
       if(!set->FindByName(files[i]->name))
 	 Sub(i--);
 }
 void FileSet::SubtractSameType(const FileSet *set)
 {
+   if(!set)
+      return;
    for(int i=0; i<fnum; i++)
    {
       FileInfo *f=set->FindByName(files[i]->name);
@@ -300,6 +320,8 @@ void FileSet::SubtractSameType(const FileSet *set)
 }
 void FileSet::SubtractDirs(const FileSet *set)
 {
+   if(!set)
+      return;
    for(int i=0; i<fnum; i++)
    {
       if(!files[i]->TypeIs(FileInfo::DIRECTORY))
@@ -311,6 +333,8 @@ void FileSet::SubtractDirs(const FileSet *set)
 }
 void FileSet::SubtractNotOlderDirs(const FileSet *set)
 {
+   if(!set)
+      return;
    for(int i=0; i<fnum; i++)
    {
       if(!files[i]->TypeIs(FileInfo::DIRECTORY)
@@ -391,19 +415,25 @@ void FileSet::ExcludeCompound()
 {
    for(int i=0; i<fnum; i++)
    {
-      if(strchr(files[i]->name,'/'))
+      const char *name=files[i]->name;
+      if(!strncmp(name,"./~",3))
+	 name+=3;
+      if(strchr(name,'/'))
 	 Sub(i--);
    }
 }
 
-void FileSet::ExcludeUnaccessible()
+void FileSet::ExcludeUnaccessible(const char *user)
 {
    for(int i=0; i<fnum; i++)
    {
-      if(!(files[i]->defined&FileInfo::MODE) || !(files[i]->defined&FileInfo::TYPE))
+      if(!files[i]->Has(FileInfo::MODE) || !files[i]->Has(FileInfo::TYPE))
 	 continue;
-      if((files[i]->filetype==FileInfo::NORMAL    && !(files[i]->mode&0444))
-      || (files[i]->filetype==FileInfo::DIRECTORY && !(files[i]->mode&0444&(files[i]->mode<<2))))
+      unsigned mask=0444;
+      if(user && files[i]->Has(FileInfo::USER))
+	 mask=(!strcmp(files[i]->user,user)?0400:0044);
+      if((files[i]->TypeIs(FileInfo::NORMAL)    && !(files[i]->mode&mask))
+      || (files[i]->TypeIs(FileInfo::DIRECTORY) && !(files[i]->mode&mask&(files[i]->mode<<2))))
       {
 	 Sub(i);
 	 i--;
@@ -476,12 +506,15 @@ void FileSet::Count(int *d,int *f,int *s,int *o) const
       switch(files[i]->filetype)
       {
       case(FileInfo::DIRECTORY):
-	 if(d) (*d)++; break;
+	 if(d) (*d)++;
+	 break;
       case(FileInfo::NORMAL):
-	 if(f) (*f)++; break;
+	 if(f) (*f)++;
+	 break;
       case(FileInfo::SYMLINK):
-	 if(s) (*s)++; break;
-      case(FileInfo::UNKNOWN):
+	 if(s) (*s)++;
+	 break;
+      default:
 	 if(o) (*o)++;
       }
    }
@@ -548,7 +581,7 @@ static bool do_exclude_match(const char *prefix,const FileInfo *fi,const Pattern
    return x->MatchExclude(name);
 }
 
-void  FileSet::Exclude(const char *prefix,const PatternSet *x)
+void  FileSet::Exclude(const char *prefix,const PatternSet *x,FileSet *fsx)
 {
    if(!x)
       return;
@@ -556,7 +589,10 @@ void  FileSet::Exclude(const char *prefix,const PatternSet *x)
    {
       if(do_exclude_match(prefix,files[i],x))
       {
-	 Sub(i);
+	 if(fsx)
+	    fsx->Add(Borrow(i));
+	 else
+	    Sub(i);
 	 i--;
       }
    }
@@ -717,12 +753,6 @@ FileInfo::FileInfo(const FileInfo &fi)
    nlinks=fi.nlinks;
    longname.set(fi.longname);
 }
-FileInfo::FileInfo(const char *n)
-{
-   Init();
-   SetName(n);
-}
-
 FileInfo::~FileInfo()
 {
 }
@@ -1001,6 +1031,7 @@ void FileInfo::MakeLongName()
    case UNKNOWN:   break;
    case DIRECTORY: filetype_c='d'; break;
    case SYMLINK:   filetype_c='l'; break;
+   case REDIRECT:  filetype_c='L'; break;
    }
    int mode1=(defined&MODE?mode:
       (filetype_c=='d'?0755:(filetype_c=='l'?0777:0644)));
